typedef int bool;

typedef enum { FATHER,PREFIX,CLASS_S,PROCESS_S,COROUTINE_S,VARIABLE_S,SIGNAL_S,
               CONSTANT_S,FUNCTION_S,PROCEDURE_S,BLOCK_S,HANDLERS_S }  type;

typedef enum { CLOSE,HIDDEN,TAKEN,NORMAL }  spec;

typedef enum { INTEGER,REAL,BOOLEAN,CHARACTER,STRING,FILE,SEMAPHORE,POINTER } var_type;

typedef enum { ALL,NOT_HIDDEN } visible;

typedef enum { SHARED,NOT_SHARED } sharing;




#define NO_OFFSET    -2
#define NO_IDENT     NULL
#define OFFSET          150   
#define TRUE          1
#define FALSE         0
#define MAIN_LEVEL    0   
#define DOT           TRUE
#define NOT_DOT       FALSE

class packet {public : };

class entry
{
public :
  String  id;
  type    kind;
  int     offset;
  spec    visible;
  packet* rest;
  entry*  prev;

  entry(String,type,int,spec,packet*);
};

struct find_result
{
  int    level;
  entry* found;         

  find_result(int l,entry* e) { level=l;found=e;}
};


class list_elem
{
public :
  entry*      info_ptr;
  list_elem*  next;
  list_elem(entry*  e) { info_ptr=e; };
};

class list
{
public :
  list_elem*  end_ptr;
  void        add_elem(entry*);

  entry*      find_virt(String); 

  list()      { end_ptr=NULL; };
};

class spec_elem
{
public :
  String      info;
  spec_elem*  next;

  spec_elem(String e,spec_elem* n) { info=e;next=n; };
};

class spec_list
{
public :
  spec_elem*  end_ptr;

  void        add    ( String );
  bool        present( String );

  spec_list() { end_ptr=NULL; };
};

class pref_list_elem
{
  public :
  entry*    where;
  int       base;
  sharing      shared;
  pref_list_elem* next; 

  pref_list_elem(entry* e,sharing sh,pref_list_elem* c)
  {
    where  = e;
    shared = sh;
    next   = c;
  };

};


class pref_list
{
public :
  pref_list_elem*  end_ptr;
  void add_pref(entry*,sharing);
  pref_list() {end_ptr=NULL;};
};

class node
{
public :
  int         level;
  int         current_offset;
  entry*      end_ptr;
  entry*      begin_ptr;
  list*       virtual_list;
  list*       formal_param_list;
  list*       ref_var_list;
  pref_list*  prefix;
  spec_list*  close;
  spec_list*  hidden;
  spec_list*  taken; 

  void        add_spec(spec,String);
  void        add_virt(entry*);
  void        add_form_param(entry*);
  void        close_unit();
  node*       open_unit();
  struct find_result*   local_find(String,visible);
  node(int);
 
};

class c_class : public packet
{
public :
  pref_list*  prefix;
  bool        f_param;
  bool        virt;  // or shared
  node*       my_symtab;

};

class f_father : public packet
{
public :
  node*   father_ptr;
  int     level;
  int     base;
  f_father(node* ff,int ll) {    father_ptr=ff;level=ll;  };
};

class symboltable
{
public :
  node*   last;
  node*   root;
  entry*   inherit(String,sharing sh = NOT_SHARED);
  entry*  insert(String,type,packet*);

  struct find_result* find_in_module(node*,String,bool,visible vis = ALL);

  struct find_result* find(String);
  symboltable()  {root=last=new node(MAIN_LEVEL);};
};

class v_variable : public packet
{
public :
  var_type  type;
  bool      f_param;
};

class s_signal : public c_class {};

class h_handler : public c_class  {}; 

class c_coroutine : public c_class { };

class p_process : public c_class {};

class b_block : public c_class {};

class p_procedure : public c_class { public: } ;

class f_function : public c_class {

public :
  type       result_type;
};

class c_constant : public packet
{
public :
  int   ptr;  //here will be link to constant value
};
